from .utils import get_class_names
from .voc import VOC
from .fat import FAT
from .open_images import OpenImages
from .ferplus import FERPlus
from .fer import FER
from .cityscapes import CityScapes
from .coco import JOINT_CONFIG
from .coco import FLIP_CONFIG
from .coco import HUMAN_JOINT_CONFIG
from .CMU_poanoptic import MINIMAL_HAND_CONFIG
from .CMU_poanoptic import IK_UNIT_LENGTH
from .CMU_poanoptic import MANOHandJoints
from .CMU_poanoptic import MPIIHandJoints
from .shapes import Shapes
from .omniglot import Omniglot
